from typing import List, Set, Tuple
from releasy.repository import CommitSet

class ContributorSet():
    def __init__(self, 
            commits: CommitSet = None, 
            previous_authors: Set[str] = None) -> None:
        self._commits = CommitSet()
        self.committers = set[str]()
        self.authors = set[str]()
        self.newcomers = set[str]()
  
        if commits:
            self._commits = commits
            self.committers = set[str](commit.committer for commit in commits)
            self.authors = set[str](commit.author for commit in commits)

            if previous_authors:
                self.newcomers = self.authors - previous_authors

    def __len__(self):
        return len(self.authors)

    def frequency(self):
        contributors_frequency = dict[str,int]()
        total_contributions = 0
        for author in (commit.author for commit in self._commits):
            contributors_frequency[author] \
                = contributors_frequency.get(author, 0) +1
            total_contributions += 1

        for contributor, contributions in contributors_frequency.items():
            contributors_frequency[contributor] \
                = int(contributors_frequency[contributor] \
                    * 100 \
                    / total_contributions)
        
        return contributors_frequency

    def top(self, top: int = None, percent: int = None) -> Tuple[str, int]:
        if top and percent:
            raise ValueError("must use top or percent argument")

        contributors = sorted(
            self.frequency().items(), 
            key=lambda t: (-t[1], t[0]))

        if top == None and percent == None:
            return contributors
        
        if top:
            return contributors[0:top]

        if percent == 100:
            return contributors
        
        contributions = 0
        for pos, (contributor, frequency) in enumerate(contributors):
            contributions += frequency
            if contributions >= percent:
                return contributors[0:pos+1] 
            
        return []
    
    def commits(self, contributors: List[str]) -> CommitSet:
        def get_contributor(contributor):
            if isinstance(contributor, tuple):
                return contributor[0]
            return contributor
        
        contributors = set(
            get_contributor(contributor) for contributor in contributors)
        
        commits = [commit for commit in self._commits
            if commit.author in contributors 
                or commit.committer in contributors]
        commits = CommitSet(commits)
        return commits



"""Semantic Miner Module

This module categorizes the releases according to their semantic:

- Main release
- Patch releases

This module also categorize releases into pre releases.
"""

from typing import Any, Tuple
from unicodedata import name
from .project import Project
from .release import Release
from .semantic import MainRelease, Patch, SReleaseSet, SemanticRelease
from .miner_base import AbstractMiner


class SemanticReleaseMiner(AbstractMiner):
    """Categorize releases into major, minor and patch"""
    def __init__(self) -> None:
        super().__init__()
        self.mreleases = SReleaseSet[MainRelease]()
        self.patches = SReleaseSet[Patch]()
        self.sreleases = SReleaseSet[SemanticRelease]()
        self.r2s = dict[Release, SemanticRelease]()
        self.versions = set[str]()
        self.project = None

    def mine(self, project: Project, *args) -> Tuple[Project, Any]:
        self.project = project
        self.releases = project.releases
        self._remove_pre_releases()
        self._mine_semantic_releases()
        self._assign_patches_to_mreleases()
        self._assign_base_releases()
        self._assign_main_base_releases()
        self._assign_prev_main_releases()
  
        self.project.main_releases = self.mreleases
        self.project.patches = self.patches
        return (self.project, [])

    def _remove_pre_releases(self):
        self.releases = [release for release in 
                                 self.releases 
                                 if not release.version.is_pre_release()]

    def _mine_semantic_releases(self) -> SReleaseSet:
        #TODO implement in ReleaseSet
        for release in sorted(self.project.releases.all(), 
                              key = lambda r: (r.time, r.version)):
            duplicated_release \
                = True if release.version.number in self.versions else False
            if release.version.is_main_release() and not duplicated_release:
                srelease = MainRelease(release)
                self.mreleases.add(srelease)
                self.versions.add(release.version.number)
            else:
                srelease = Patch(release)
                self.patches.add(srelease)  
                if duplicated_release:
                    srelease.was_main_release = True

            self.r2s[release] = srelease
            self.sreleases = SReleaseSet(self.mreleases | self.patches)
    
    def _assign_patches_to_mreleases(self):
        for patch in self.patches:
            main_release = self._track_patch_main_release(patch)
            if main_release:
                main_release.patches.add(patch)
                patch.main_release = main_release
            elif patch.commits:
                patch.is_orphan = True

    def _track_patch_main_release(self, patch: Patch):
        patch_to_track = [patch]
        loop_control = set()
        while patch_to_track:
            patch = patch_to_track.pop()
            if patch.release.base_release:
                sbase_release = self.r2s[patch.release.base_release]
                if isinstance(sbase_release, MainRelease):
                    return sbase_release
                elif sbase_release not in loop_control:
                    loop_control.add(patch)
                    patch_to_track.append(sbase_release)
        return None

    def _assign_base_releases(self):
        for srelease in self.sreleases:
            if srelease.release.base_release:
                srelease.base_release = self.r2s[srelease.release.base_release]
            for base_release in srelease.release.base_releases:
                srelease.base_releases.add(self.r2s[base_release])

    def _assign_main_base_releases(self):
        for srelease in self.sreleases:
            self._assign_release_main_base_release(srelease)
    
    def _assign_release_main_base_release(self, release: SemanticRelease) -> None:
        main_base_releases = SReleaseSet[SemanticRelease]()
        for srelease in release.base_releases:
            if isinstance(srelease, MainRelease):
                main_base_releases.add(srelease)
            else:
                main_base_releases.add(srelease.main_release)
                
        if len(main_base_releases) == 1:
            release.main_base_release = main_base_releases[0]
        elif len(main_base_releases) > 1:
            releases = [release]
            releases.extend(main_base_releases)
            releases = sorted(releases, key = lambda r: r.release.version)
            pos = releases.index(release)
            if pos > 0:
                release.main_base_release = releases[pos-1]
            else:
                release.main_base_release = releases[1]

    def _assign_prev_main_releases(self):        
        mreleases = sorted(self.mreleases, 
            key=lambda r: (r.release.version, r.time))
        for i in range(1, len(mreleases)):
            mrelease = mreleases[i]
            prev_mrelease_pos = i-1

            while prev_mrelease_pos >= 0 \
                    and mreleases[prev_mrelease_pos].time > mrelease.time:
                prev_mrelease_pos -= 1
            
            if prev_mrelease_pos >= 0:
                prev_mrelease = mreleases[prev_mrelease_pos]
                mrelease.prev_main_release = prev_mrelease





from typing import Any, Tuple
from releasy.miner_base import AbstractMiner
from releasy.project import Project
from releasy.contributor import ContributorSet


class ContributorMiner(AbstractMiner):
    def __init__(self) -> None:
        super().__init__()

    def mine(self, project: Project, *args) -> Tuple[Project, Any]:
        previous_authors = set[str]()
        for release in sorted(
                project.releases, key = lambda r: (r.time, r.version)):

            release.contributors = ContributorSet(
                release.commits, 
                previous_authors
            )
            previous_authors.update(release.contributors.authors)
        
        return project, []



# This module connect releasy with Git

from datetime import datetime, timedelta, timezone
from typing import Dict, List, Set, Tuple
import pygit2

from releasy.repository import Commit, CommitSet, DiffDelta, Repository, RepositoryProxy, Tag


class GitRepository(RepositoryProxy):
    """
    A Repository proxy to Git
    """
    def __init__(self, path) -> None:
        super().__init__()
        self.path = path
        self.name = path
        self.git: pygit2.Repository = pygit2.Repository(path)
        self.commit_cache = CommitCache(self.git)
        self.repository: Repository = None

    def fetch_tags(self) -> Set[Tag]:
        rtags = [ref for ref in self.git.references.objects 
                    if ref.name.startswith('refs/tags/')]

        tags: Set[Tag] = set()
        for rtag in rtags:
            tag = self._get_tag(rtag)
            if tag:
                tags.add(tag)
        return tags

    def _get_tag(self, rtag: pygit2.Reference) -> Tag:
        ref = self.git.get(rtag.target)
        if ref.type == pygit2.GIT_OBJ_COMMIT:
            commit = self.repository.get_commit(ref.hex)
            #TODO time
            tag = Tag(self.repository, rtag.shorthand, commit)
            return tag                
        elif ref.type == pygit2.GIT_OBJ_TAG: # annotatted tag
            peel = rtag.peel()
            if peel.type == pygit2.GIT_OBJ_COMMIT:
                commit = self.repository.get_commit(rtag.peel().hex)
                rtag_ref: pygit2.Tag = ref
                try:
                    message = rtag_ref.message
                except:
                    message = ''
                if rtag_ref.tagger:
                    tagger = f"{rtag_ref.tagger.name} <{rtag_ref.tagger.email}>"
                    time_tzinfo = timezone(timedelta(minutes=rtag_ref.tagger.offset))
                    time = datetime.fromtimestamp(float(rtag_ref.tagger.time), time_tzinfo)
                    tag = Tag(self.repository, rtag.shorthand, commit, message, tagger, 
                            time)
                else:
                    tag = Tag(self.repository, rtag.shorthand, commit, message)
                return tag
        return None

    def fetch_commit(self, commit_id: str) -> Commit:
        rcommit = self.commit_cache.fetch_commit(commit_id)

        committer_tzinfo = timezone(timedelta(minutes=rcommit.committer.offset))
        committer_time = datetime.fromtimestamp(float(rcommit.committer.time), committer_tzinfo)
        author_tzinfo = timezone(timedelta(minutes=rcommit.author.offset))
        author_time = datetime.fromtimestamp(float(rcommit.author.time), author_tzinfo)

        try:
            message = rcommit.name,
        except:
            message = ''

        commit = Commit(
            self.repository,
            rcommit.hex,
            message,
            f"{rcommit.committer.name} <{rcommit.committer.email}>",
            committer_time,
            f"{rcommit.author.name} <{rcommit.author.email}>",
            author_time)
        return commit

    def fetch_commit_parents(self, commit: Commit) -> CommitSet:
        commit_ref: pygit2.Commit = self.commit_cache.fetch_commit(commit.id)
        parents = CommitSet()
        for parent_ref in commit_ref.parents:
            parent = self.repository.get_commit(parent_ref.hex)
            parents.add(parent)
        return parents

    def diff(self, commit_a: Commit, commit_b: Commit, parse_delta:bool = False) -> DiffDelta:
        diff_result = self.git.diff(commit_a.id, commit_b.id)

        files = set()
        if parse_delta:
            for delta in diff_result.deltas:
                if delta.new_file.path:
                    files.add(delta.new_file.path)
                if delta.old_file.path:
                    files.add(delta.old_file.path)

        delta = DiffDelta(
            diff_result.stats.insertions, 
            diff_result.stats.deletions,
            diff_result.stats.files_changed,
            files)
        return delta

class CommitCache:
    """
    Implement a cache to improve fech commit performance
    """
    def __init__(self, git: pygit2.Repository) -> None:
        self.git =  git
        self.cache: Dict[str, pygit2.Commit] = {}

    def fetch_commit(self, commit_id: str) -> pygit2.Commit:
        if commit_id not in self.cache:
            commit_ref: pygit2.Commit = self.git.get(commit_id)
            self.cache[commit_id] = commit_ref
        return self.cache[commit_id]



from __future__ import annotations
from ast import List
from collections import OrderedDict
from typing import Callable, Dict, Generic, Iterator, Set, TypeVar
from typing import TYPE_CHECKING

from releasy.contributor import ContributorSet
if TYPE_CHECKING:
    from releasy.project import Project

import re
from datetime import datetime, timedelta

from .repository import (
    Commit,
    CommitSet, 
    Repository, 
    Tag)


class Release:
    """Represent a software release.

    Attributes:
      - name: the release name
      - version: the release version
      - tag: the release tag
      - commits: all the release commits
      - head: the last commit of a release
      - tails: the first commits of a release
      - base_releases: the imediatelly reacheable releases
    """
    def __init__(self, project: Project, name: str, tag: Tag) -> None:
        self.project = project
        self.name = name
        self.tag = tag
        if tag: #Fix must remove tag from release
            self.head = tag.commit
        self.commits = CommitSet()
        self.tails = set[Commit]()
        self.base_release: Release = None
        self.base_releases = ReleaseSet()
        self.version = ReleaseVersion(name)
        self.contributors: ContributorSet = ContributorSet()
    
    def __hash__(self):
        return hash((self.project, self.name))

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, Release):
            return self.project == __o.project and self.name == __o.name
        else:
            return False

    def __repr__(self) -> str:
        return self.name

    @property
    def time(self) -> datetime:
        return self.tag.time

    @property
    def merges(self):
        return CommitSet(
            commit for commit in self.commits if len(commit.parents) > 1)

    # @property
    # def cycle(self) -> timedelta:
    #     return 

class ReleaseName():
    pass

TYPE_MAJOR   = 0b10000
TYPE_MINOR   = 0b01000
TYPE_MAIN    = 0b00100
TYPE_PRE     = 0b00010
TYPE_PATCH   = 0b00001
class ReleaseVersion():
    def __init__(self, name: str, separator: re.Pattern = None,
                 version_separator: re.Pattern = None) -> None:
        self.full_name = name
        self.prefix = ''
        self.suffix = ''
        
        if not separator:
            separator = re.compile(
                r'(?P<prefix>(?:[^\s,]*?)(?=(?:[0-9]+[\._]))|[^\s,]*?)(?P<version>(?:[0-9]+[\._])*[0-9]+)(?P<suffix>[^\s,]*)'
            )
        parts = separator.match(name)
        if parts:
            if parts.group('prefix'):
                self.prefix = parts.group('prefix')
            if parts.group('suffix'):
                self.suffix = parts.group('suffix')
            if parts.group('version'):
                self.number = parts.group('version')

        if not version_separator:
            version_separator = re.compile(r'([0-9]+)')
        version_parts = version_separator.findall(self.number)
        self.numbers = [int(version_part) for version_part in version_parts]
        if len(self.numbers) == 1:
            self.numbers.append(0)
        if len(self.numbers) == 2:
            self.numbers.append(0)
        #TODO handle dinamic
        self.number = '.'.join([str(v) for v in self.numbers])

    def __lt__(self, other):
        return self.__cmp(self, other) < 0
    def __gt__(self, other):
        return self.__cmp(self, other) > 0
    def __eq__(self, other):
        return self.__cmp(self, other) == 0
    def __le__(self, other):
        return self.__cmp(self, other) <= 0
    def __ge__(self, other):
        return self.__cmp(self, other) >= 0

    def __cmp(self, version_a: ReleaseVersion, version_b: ReleaseVersion) -> int:
        for vnumber_a, vnumber_b in zip(version_a.numbers, version_b.numbers):
            if vnumber_a > vnumber_b:
                return 1
            if vnumber_a < vnumber_b:
                return -1
        if not version_a.suffix and version_b.suffix:
            return 1
        elif version_a.suffix and not version_b.suffix:
            return -1

        if version_a.suffix and version_b.suffix:
            if version_a.suffix > version_b.suffix:
                return 1
            elif version_a.suffix < version_b.suffix:
                return -1

        return 0

    def is_main_release(self) -> bool:
        return self.type(TYPE_MAIN)

    def is_pre_release(self) -> bool:
        return self.type(TYPE_PRE)
    
    def is_major(self) -> bool:
        return self.type(TYPE_MAJOR)

    def is_minor(self) -> bool:
        return self.type(TYPE_MINOR)

    def is_patch(self) -> bool:
        return self.type(TYPE_PATCH)

    def type(self, mask: int) -> bool:
        type = 0b0
        if int(self.numbers[2]) > 0:
            type |= TYPE_PATCH
        elif int(self.numbers[1]) > 0:
            type |= TYPE_MINOR
        elif int(self.numbers[0]) > 0:
            type |= TYPE_MAJOR

        if self.suffix:
            type |= TYPE_PRE
        elif not (type & TYPE_PATCH):
            type |= TYPE_MAIN

        if type & mask == mask:
            return True
        else:
            return False

    #TODO return ReleaseVersion
    def normalize(self, n: int):
        """ Return a version str with size n. If n is greater than the actual
        version len, it append [0] to the right."""
        size = len(self.numbers)
        if size <= n:
            return self.numbers + [0]*(n-size)
        else:
            return self.numbers[0:n]

    def diff(self, target: ReleaseVersion):
        """
        Compare two versions and return a vector
        Ex: 2.0.0 - 1.0.1 = [1, 0, -1]
        """
        size = max(len(target.numbers), len(self.numbers))
        me = self.normalize(size)
        target = target.normalize(size)
        result = []
        for v1,v2 in zip(me, target):
            result.append(v1 - v2)
        return result
    

class ReleaseSet():
    def __init__(self, releases: Set[Release] = None) -> None:
        self._releases = OrderedDict[str, Release]()
        if releases:
            for release in releases:
                self.add(release)

    def __iter__(self) -> Iterator[Release]:
        return (release for release in self._releases.values())

    def __getitem__(self, key) -> Release:
        if isinstance(key, int):
            release_name = list(self._releases.keys())[key]
            return self._releases[release_name]
        elif isinstance(key, str):
            return self._releases[key]
        else:
            raise TypeError()

    def __contains__(self, item) -> bool:
        if isinstance(item, str) and item in self._releases:
            return True
        elif isinstance(item, Release) and item.name in self._releases:
            return True
        return False

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, ReleaseSet):
            return self._releases == __o._releases
        return False

    @property
    def names(self) -> Set[str]:
        """Return a set with all release names"""
        return set(name for name in self._releases.keys())

    def prefixes(self) -> Set[str]:
        """Return a set with all the release prefixes"""
        prefixes = set[str]()
        for release in self._releases.values():
            prefixes.add(release.version.prefix)
        return prefixes

    def first(self, func: Callable = None) -> Release:
        if self._releases:
            if func:
                ordered_releases = sorted(self._releases.values(), key=func)
            else:
                ordered_releases = list(self._releases.values())
            return ordered_releases[0]
        else:
            return None

    def last(self, func: Callable = None) -> Release:
        if self._releases:
            if func:
                ordered_releases = sorted(self._releases.values(), key=func)
            else:
                ordered_releases = list(self._releases.values())
            return ordered_releases[-1]
        else:
            return None

    def all(self) -> List[Release]:
        return list(self._releases.values())

    def add(self, release: Release):
        if release:
            self._releases[release.name] = release

    def remove(self, release: Release):
        if release.name in self:
            del self._releases[release.name]

    def update(self, iterable):
        for item in iterable:
            self.add(item)

    def __len__(self):
        return len(self._releases)


class Commit2ReleaseMapper():
    """Map a commit to a release"""
    def __init__(self) -> None:
        self.mapper = dict[Commit, ReleaseSet]()

    def get_release(self, commit: Commit) -> bool:
        return self.mapper.get(commit)

    def assign_commit(self, commit: Commit, release: Release):
        if commit not in self.mapper:
            self.mapper[commit] = set() 
        self.mapper.get(commit).add(release)


class BaseReleaseSet():
    """"""
    def __init__(self, release: Release, base_releases: ReleaseSet) -> None:
        self.release = release
        self.base_releases = base_releases



# This module abstract repository objects such as commits and tags
from __future__ import annotations
from abc import ABC, abstractclassmethod
from collections import OrderedDict
from mimetypes import init
from typing import Callable, Dict, Iterator, Set
from datetime import datetime


class Repository:
    """ 
    A repository stores Tags and Commits
    """
    def __init__(self, proxy: RepositoryProxy):
        proxy.repository = self # TODO better alternative
        self.name: str = proxy.name
        self.proxy = proxy
        self.commit_cache = CommitCache(proxy)

    def get_tags(self) -> Set[Tag]:
        tags = self.proxy.fetch_tags()
        return tags

    def get_commit(self, id: str) -> Commit:
        commit = self.commit_cache.fetch_commit(id)
        return commit

    def get_parents(self, commit: Commit) -> CommitSet:
        parents = self.proxy.fetch_commit_parents(commit)
        return parents

    def diff(self, commit_a: Commit, commit_b: Commit, parse_files: bool) -> DiffDelta:
        diff_delta = self.proxy.diff(commit_a, commit_b, parse_files)
        return diff_delta


class CommitCache:
    """
    Implement cache to improve fech commit performance
    """
    def __init__(self, proxy: RepositoryProxy) -> None:
        self.cache: Dict[str, Commit] = {}
        self.proxy = proxy

    def fetch_commit(self, commit_id: str) -> Commit:
        if commit_id not in self.cache:
            commit = self.proxy.fetch_commit(commit_id)
            self.cache[commit_id] = commit
        return self.cache[commit_id]


class RepositoryProxy(ABC):
    """ 
    An adapter to enable developers creating specific repository, such as Git 
    """
    def __init__(self) -> None:
        super().__init__()
        self.repository: Repository = None #TODO better alternative
        self.name: str = None

    @abstractclassmethod
    def fetch_tags(self) -> Set[Tag]:
        pass

    @abstractclassmethod
    def fetch_commit(self, id: str) -> Commit:
        pass

    @abstractclassmethod
    def fetch_commit_parents(self, commit: Commit) -> CommitSet:
        pass

    @abstractclassmethod
    def diff(self, commit_a: Commit, commit_b: Commit, parse_files:bool) -> DiffDelta:
        pass


class Tag:
    """
    A Tag represents a reference to a commit, which is a potential release
    """
    def __init__(self, repository: Repository, name: str, commit: Commit = None,
                 message: str = None, author: str = None, time: datetime = None) -> None:
        self.repository = repository
        self.name = name
        self.commit = commit
        self.author = author
        if message:
            self.message = message
        else:
            self.message = ""
        if time:
            self.time = time
        elif commit:
            self.time = commit.committer_time
        if message or time:
            self.is_annotated = True
        else:
            self.is_annotated = False

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, Tag):
            return self.repository == __o.repository and self.name == __o.name
        else:
            return False

    def __hash__(self) -> int:
        return hash((self.repository, self.name))

    def __repr__(self) -> str:
        return self.name


class Commit:
    """
    A Commit represents a change
    """
    def __init__(self, repository: Repository, id: str, message: str = None, 
                 committer: str = None, committer_time: datetime = None,
                 author: str = None, author_time: datetime = None) -> None:
        self.repository = repository
        self.id = id
        self.message = message
        self._parents = None # Lazy loaded
        self.committer = committer
        self.committer_time = committer_time
        self.author = author
        self.author_time = author_time

    @property
    def parents(self) -> CommitSet:
        if not self._parents:
            self._parents = self.repository.get_parents(self)
        return self._parents

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, Commit):
            return self.repository == __o.repository and self.id == __o.id
        else:
            return False
        
    def __hash__(self) -> int:
        return hash((self.repository, self.id))

    def __repr__(self) -> str:
        return self.id[0:8]

    def diff(self, commit: Commit, parse_files:bool = False) -> DiffDelta:
        delta = self.repository.diff(self, commit, parse_files)
        return delta
    
    # TODO commit.history_until(release_commits)
    # for commit in commits:
    #     release_commits = commit.history_until(release_commits)


class DiffDelta:
    def __init__(self, insertions: int, deletions: int, files_changed: int, 
            files: Set[str]):
        self.insertions = insertions
        self.deletions = deletions
        self.files_changed = files_changed
        self.files = files
   
    @property
    def churn(self) -> int:
        return self.insertions + self.deletions


class CommitSet:
    def __init__(self, commits: Set[Commit] = None) -> None:
        self._commits = OrderedDict[str, Commit]()
        if commits:
            self.update(commits)
    
    def __contains__(self, item) -> bool:
        if isinstance(item, str) and item in self._commits:
            return True
        elif isinstance(item, Commit) and item.id in self._commits:
            return True
        return False

    def __getitem__(self, key) -> Commit:
        if isinstance(key, int):
            commit_id = list(self._commits.keys())[key]
            return self._commits[commit_id]
        elif isinstance(key, str):
            return self._commits[key]
        else:
            raise TypeError()

    def __iter__(self) -> Iterator[Commit]:
        return iter(self._commits.values())

    def __eq__(self, __o: object) -> bool:
        return self.all == __o

    def __len__(self) -> int:
        return len(self._commits)
        
    def add(self, commit: Commit):
        if commit and commit not in self._commits:
            self._commits[commit.id] = commit
    
    def update(self, commits):
        for commit in commits:
            self.add(commit)

    def __sub__(self, other):
        return self.all - other.all

    @property
    def ids(self) -> Set[str]:
        return set(commit.id for commit in self._commits.values())

    @property
    def all(self) -> Set[Commit]:
        return set(commit for commit in self._commits.values())

    @property
    def authors(self) -> Set[str]:
        return set(commit.author for commit in self._commits.values())

    @property
    def committers(self) -> Set[str]:
        return set(commit.committer for commit in self._commits.values())

    def first(self, func: Callable = None) -> Commit:
        if self._commits:
            if func:
                ordered_commits = sorted(self._commits.values(), key=func)
            else:
                ordered_commits = list(self._commits.values())
            return ordered_commits[0]
        else:
            return None

    def last(self, func: Callable = None) -> Commit:
        if self._commits:
            if func:
                ordered_commits = sorted(self._commits.values(), key=func)
            else:
                ordered_commits = list(self._commits.values())
            return ordered_commits[-1]
        else:
            return None 


from __future__ import annotations
from collections import OrderedDict

from typing import Callable, Generic, Iterator, Set, TypeVar

import datetime

from .project import Project
from .release import Release, ReleaseSet
from .repository import CommitSet


class SemanticRelease:
    def __init__(self, release: Release):
        self.project = release.project #TODO may be removed?
        self.release = release
        self.main_base_release: MainRelease = None
        self.base_release: SemanticRelease = None
        self.base_releases: SReleaseSet = SReleaseSet()
        self.prev_main_release = None
        
    @property
    def name(self):
        return self.release.name

    @property
    def commits(self):
        return self.release.commits
    
    def __hash__(self):
        return hash((self.project, self.name))

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, SemanticRelease):
            return self.project == __o.project and self.name == __o.name
        else:
            return False

    def __repr__(self) -> str:
        return self.name

    # def __lt__(self, other):
    #     return self.release.version < other.release.version < 0
    # def __gt__(self, other):
    #     return self.release.version > other.release.version
    # def __le__(self, other):
    #     return self.release.version <= other.release.version
    # def __ge__(self, other):
    #     return self.release.version >= other.release.version

    @property
    def time(self) -> datetime.datetime:
        return self.release.time


class FinalRelease(SemanticRelease):
    pass

#TODO implement pre release logic
class PreRelease(SemanticRelease):
    pass


class MainRelease(SemanticRelease):
    def __init__(self, release: Release) -> None:
        super().__init__(release)
        self.patches = SReleaseSet()

    @property
    def cycle(self) -> datetime.timedelta:
        if self.prev_main_release:
            ref = self.prev_main_release.time
        elif self.commits:
            ref = self.commits.first(lambda c: c.author_time).author_time
        else:
            return datetime.timedelta(0)
        return self.time - ref

    @property
    def delay(self) -> datetime.timedelta:
        if self.prev_main_release and self.commits:
            ref = self.release.commits.first(
                lambda c: c.author_time).author_time
            return ref - self.prev_main_release.time
        else:
            return datetime.timedelta(0)


class Patch(FinalRelease):
    def __init__(self, release: Release) -> None:
        super().__init__(release)   
        self.main_release: MainRelease = None
        self.is_orphan = False
        self.was_main_release = False

    @property
    def cycle(self) -> datetime.timedelta:
        if self.base_release:
            ref = self.base_release.time
        else:
            ref = self.commits.first(lambda c: c.author_time).author_time
        return self.time - ref

    @property
    def delay(self) -> datetime.timedelta:
        if self.base_release:
            ref = self.release.commits.first(
                lambda c: c.author_time).author_time
            return ref - self.base_release.time
        else:
            return datetime.timedelta(0)


SR = TypeVar('SR')
class SReleaseSet(Generic[SR]):
    def __init__(self, sreleases: Set[SR] = None) -> None:
        self._sreleases = OrderedDict[str, SR]()
        if sreleases:
            for srelease in sreleases:
                self.add(srelease)

    def __len__(self) -> int:
        return len(self._sreleases)

    def __iter__(self) -> Iterator[SR]:
        return iter(self._sreleases.values())

    def __contains__(self, item) -> bool:
        if isinstance(item, str) and item in self._sreleases:
            return True
        elif isinstance(item, SemanticRelease) and item.name in self._sreleases:
            return True
        else:
            return False

    def __getitem__(self, key) -> SR:
        if isinstance(key, int):
            release_name = list(self._sreleases.keys())[key]
            return self._sreleases[release_name]
        elif isinstance(key, str):
            return self._sreleases[key]
        else:
            raise TypeError()

    def __repr__(self) -> str:
        return str(set(self._sreleases.keys()))

    def commits(self) -> CommitSet():
        commits = CommitSet()
        for srelease in self._sreleases.values():
            commits.update(srelease.commits)
        return commits

    @property
    def names(self) -> Set[str]:
        """Return a set with all release names"""
        return set(name for name in self._sreleases.keys())

    @property
    def all(self) -> Set[SR]:
        return set(self._sreleases.values())

    def prefixes(self) -> Set[str]:
        """Return a set with all the release prefixes"""
        prefixes = set[str]()
        for srelease in self._sreleases.values():
            for prefix in srelease.releases.prefixes():
                prefixes.add(prefix)
        return prefixes

    def first(self, func: Callable = None) -> SR:
        if self._sreleases:
            if func:
                ordered_sreleases = sorted(self._sreleases.values(), key=func)
            else:
                ordered_sreleases = list(self._sreleases.values())
            return ordered_sreleases[0]
        else:
            return None

    def last(self, func: Callable = None) -> SR:
        if self._sreleases:
            if func:
                ordered_sreleases = sorted(self._sreleases.values(), key=func)
            else:
                ordered_sreleases = list(self._sreleases.values())
            return ordered_sreleases[-1]
        else:
            return None

    def __or__(self, __o: SReleaseSet[SR]) -> SReleaseSet[SR]:
        sreleases = SReleaseSet[SR]()
        sreleases.update(self.all)
        sreleases.update(__o.all)
        return sreleases

    def add(self, srelease: SR) -> None:
        if srelease and srelease.name not in self._sreleases:
            self._sreleases[srelease.name] = srelease

    def update(self, sreleases: SR) -> None:
        for srelease in sreleases:
            self.add(srelease)
    




from __future__ import annotations

__all__ = [
    'Miner',
    'Project',
    'ReleaseMiner',
    'FinalReleaseMiner',
    'HistoryCommitMiner',
    'MixedHistoryCommitMiner',
    'ContributorMiner',
    'BaseReleaseMiner',
    'SemanticReleaseMiner']

from .miner_release import *
from .miner_commit import *
from .miner_base_release import *
from .miner_contributor import *
from .miner_semantic import *

from .project import Project
from .repository import Repository
from .miner_base import AbstractMiner
from .repository_git import GitRepository


class Miner():
    """Wrapper to Releasy Miner implementations, e.g:
   
    project = releasy.Miner('repos/releasy').apply(
        FinalReleaseMiner(),
        HistoryCommitMiner(),
        BaseReleaseMiner(),
        SemanticReleaseMiner()
    ).mine()
    """
    def __init__(self, repository: Repository, name: str = None) -> None:
        if isinstance(repository, str):
            self.repository = Repository(GitRepository(repository))
        else:
            self.repository = repository
        if name:
            self.name = name
        else:
            self.name = self.repository.name
        self.miners = list[AbstractMiner]()

    def apply(self, *miners) -> Miner:
        self.miners.extend(miners)
        return self

    def mine(self) -> Project:
        project = Project(self.name, self.repository)
        args = []
        for miner in self.miners:
            project, args = miner.mine(project, *args)
        return project



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Dict, Set, Tuple

from releasy.project import Project

from .repository import Repository, Tag


class AbstractMiner(ABC):
    def __init__(self) -> None:
        self.project: Project = None

    @abstractmethod
    def mine(self, project: Project, *args) -> Tuple[Project, Any]:
        pass



"""Commit Miner

This module mine commits and assign them to releases.

For each release, it assigns:
- its commits: the commits that belong to the release; and
- its tail commits: the first commits of each release branch
"""

from typing import Any, List, Tuple
from releasy.release import Commit2ReleaseMapper, Release

from releasy.repository import Commit, CommitSet
from .miner_base import AbstractMiner
from .project import Project


class MixedHistoryCommitMiner(AbstractMiner):
    def __init__(self) -> None:
        super().__init__()
        self.c2r = dict[Commit, Release]()

    def mine(self, project: Project, *args) -> Tuple[Project, Any]:
        self.project = project
        self._mine_commits()
        return (self.project, [self.c2r])

    def _mine_commits(self) -> None:
        release_commits = set(map(lambda release: release.tag.commit, 
                              self.project.releases))

        for release in self.project.releases:
            commits = CommitSet()
            tails = CommitSet()
            loop_detector = set()
            commits_to_track: List[Commit] = [release.head]

            if release.head not in self.c2r:
                self.c2r[release.head] = set()
            self.c2r[release.head].add(release)

            while commits_to_track:
                commit = commits_to_track.pop()
                commits.add(commit)
                loop_detector.add(commit)

                if commit.parents:
                    for parent in commit.parents:
                        if parent not in release_commits:
                            if parent not in loop_detector:
                                commits_to_track.append(parent)
                        else:
                            tails.add(commit)
                else:
                    tails.add(commit)

            release.tails = tails
            release.commits = commits


class HistoryCommitMiner(AbstractMiner):
    
    def __init__(self) -> None:
        super().__init__()
        self.c2r = Commit2ReleaseMapper()

    def mine(self, project: Project, *args) -> Tuple[Project, Any]:
        self.project = project
        self._assign_heads()
        self._mine_releases()
        return (self.project, [self.c2r]) 

    def _assign_heads(self):
        for release in sorted(
                self.project.releases, 
                key=lambda r: (r.time, r.version)):
            if not self.c2r.get_release(release.head):
                self.c2r.assign_commit(release.head, release)

    def _mine_releases(self) -> None:
        for release in sorted(
                self.project.releases, 
                key=lambda r: (r.time, r.version)):
            commits, tails = self._mine_commits(release)
            release.commits = commits
            release.tails = tails

    def _mine_commits(self, release):
        commit = release.head

        commit_release = self.c2r.get_release(commit) 
        if commit_release and release not in commit_release:
            return CommitSet(), CommitSet()
        
        commits = CommitSet()
        tails = CommitSet()
        commits_to_track: List[Commit] = [commit]    
        while commits_to_track:
            commit = commits_to_track.pop()
            
            self.c2r.assign_commit(commit, release)
            commits.add(commit)

            if not commit.parents:
                tails.add(commit)

            for parent in commit.parents:
                if not self.c2r.get_release(parent):
                    commits_to_track.append(parent)
                else:
                    tails.add(commit)
                    
        return commits, tails
    



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .repository import Repository
    from .semantic import SReleaseSet, MainRelease, Patch
    from .release import ReleaseSet

class Project:
    def __init__(self, name: str, repository: Repository) -> None:
        self.repository = repository
        self.main_releases: SReleaseSet[MainRelease] = None
        self.patches: SReleaseSet[Patch] = None
        self.releases: ReleaseSet = None
        self.name = name



"""Base Release Miner

This module mine release base release, that is the release that was created 
before the actual release.
"""
from __future__ import annotations
from typing import Any, Tuple
from releasy.miner_base import AbstractMiner
from releasy.release import Commit2ReleaseMapper, Release, ReleaseSet
from releasy.project import Project
from releasy.repository import Commit


class BaseReleaseMiner(AbstractMiner):
    def mine(self, project: Project, *args) -> Tuple[Project, Any]:
        self.project = project

        self.c2r = Commit2ReleaseMapper()
        for arg in args:
            if isinstance(arg, Commit2ReleaseMapper):
                self.c2r = arg
        self._mine_base_releases()

        return self.project, []

    def _mine_base_releases(self) -> None:
        for release in self.project.releases:
            self._mine_release_base_releases(release)
            self._mine_release_main_base_release(release)

    def _mine_release_base_releases(self, release: Release):
        base_releases = ReleaseSet(
            release
            for tail in release.tails 
            for parent in tail.parents 
            for release in self.c2r.get_release(parent))

        base_releases.remove(release)
        release.base_releases = base_releases
    
    def _mine_release_main_base_release(self, release: Release):
        if len(release.base_releases) == 1:
            release.base_release = release.base_releases[0]
        elif len(release.base_releases) > 1:
            #TODO handle v2.0.0, v2.0.1, and 2,0
            releases = [release]
            releases.extend(release.base_releases)
            releases = sorted(releases, key = lambda r: r.version)
            pos = releases.index(release)
            if pos > 0:
                release.base_release = releases[pos-1]
            else:
                release.base_release = releases[1]



from ensurepip import version
from typing import Any, Set, Tuple
import re

from .project import Project
from .release import Release, ReleaseSet
from .miner_base import AbstractMiner
from .repository import Repository

class ReleaseMiner(AbstractMiner):
    """ 
    Mine all releases 
    """
    def __init__(self) -> None:
        super().__init__()
        self.release_regexp = re.compile(
            r'(?P<prefix>(?:[^\s,]*?)(?=(?:[0-9]+[\._]))|[^\s,]*?)(?P<version>(?:[0-9]+[\._])*[0-9]+)(?P<suffix>[^\s,]*)'
        )

    def mine(self, project: Project, *args) -> Tuple[Project, Any]:
        tags = project.repository.get_tags()

        releases: Set[Release] = set()
        for tag in tags:
            if self.release_regexp.match(tag.name):
                release = Release(project, tag.name, tag)
                releases.add(release)

        project.releases = ReleaseSet(releases)
        return (project, [])

class FinalReleaseMiner(ReleaseMiner):
    """ 
    Mine only final releases, ignoring pre releases 
    """
    def mine(self, project: Project, *args) -> Tuple[Project, Any]:
        super().mine(project, args)
        releases = ReleaseSet(release for release in project.releases 
                              if not release.version.is_pre_release())
        project.releases = releases
        return (project, [])


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['releasy']

package_data = \
{'': ['*']}

install_requires = \
['pygit2>=1.11.1,<2.0.0', 'python-dateutil>=2.8.2,<3.0.0', 'pyyaml>=6.0,<7.0']

setup_kwargs = {
    'name': 'releasy-lib',
    'version': '4.2.2',
    'description': 'A library to mine release data',
    'long_description': 'Releasy\n=======\n\n[![Build Status](https://github.com/gems-uff/releasy/actions/workflows/pytest.yml/badge.svg)](https://github.com/gems-uff/releasy/actions/workflows/pytest.yml)\n\nReleasy is a tool that collects provenance data from releases \nby parsing the software version control and issue tracking\nsystems.\n\nPapers\n======\n\n[Curty, F., Kohwalter, T., Braganholo, V., Murta, L., 2018. An Infrastructure for Software Release Analysis through Provenance Graphs. Presented at the VI Workshop on Software Visualization, Evolution and Maintenance.](https://goo.gl/9u8rzc)\n\n',
    'author': 'Felipe Curty',
    'author_email': 'felipecrp@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gems-uff/releasy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
